// News.cpp: implementation of the CNews class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "News.h"
#include "textworks.h"
#include "label.h"
#include "lsapiex.h"

unsigned __stdcall UpdateNews(void* lpParameter);

extern UINT_PTR LastEventIDValue;
extern HWND messageHandler;

void ErrorBox(const char *szMessage, bool bSilent);

////////////////////////////////////////////////////////////////////////////
//
// Constructor
// 
CNews::CNews(string sName)
{
	m_sName = sName;
	m_UpdateThread = 0;
	m_bAbort = false;
}

////////////////////////////////////////////////////////////////////////////
//
// Destructor: remove updating thread
// 
CNews::~CNews()
{
	if (m_bActive)	Disable();
	m_bAbort = true;
	while (m_UpdateThread) Sleep(50);
}

////////////////////////////////////////////////////////////////////////////
//
// Read settings from .RC. Nothing else
// 
bool CNews::ReadSettings()
{
	m_sURL				= GetPrefixedRCLine(m_sName, "URL", "http://www.shellfront.org/");

	m_UpdateInterval	= GetRCInt(NAMED_OPTION(m_sName, "UpdateInterval"), 600);

	m_sInputString		= str_replace(GetPrefixedRCString(m_sName, "InputString",	"{%}"),		"{%quote}", "\"");
	m_sInputString		= str_replace(m_sInputString, "{%dollar}", "$");
	m_sOutputString		= str_replace(GetPrefixedRCString(m_sName, "OutputString",	"{%1}"),	"{%quote}", "\"");
	m_sOutputString		= str_replace(m_sOutputString, "{%dollar}", "$");

	m_sOutputLabel		= GetPrefixedRCString(m_sName, "DisplayOn",	"");

	m_sNewLine			= GetPrefixedRCString(m_sName, "NewLine",			"\\n");
	m_sInputNewLineAs	= GetPrefixedRCString(m_sName, "DisplayNewLineAs",	" ");
	m_sNewFileLine		= GetPrefixedRCString(m_sName, "FileNewLine",		"\\n");
	m_sProgress			= GetPrefixedRCString(m_sName, "ProgressInfo",		"Loading... {%}% complete");

	//m_sUserName			= GetPrefixedRCString(m_sName, "UserName",			"");
	//m_sPassword			= GetPrefixedRCString(m_sName, "Password",			"");

	m_sLocalFile		= GetPrefixedRCLine(m_sName, "LocalFile",	"");
	m_sOutputFile		= GetPrefixedRCLine(m_sName, "OutputFile",	"");

	m_sOnChecked		= GetPrefixedRCLine(m_sName, "OnChecked",	"!none");
	m_sOnFailure		= GetPrefixedRCLine(m_sName, "OnFailure",	"!none");
	m_sOnUpdated		= GetPrefixedRCLine(m_sName, "OnUpdated",	"!none");

	m_bActive			= GetPrefixedRCBool(m_sName, "Enabled",		true);
	m_bShowProgress		= GetPrefixedRCBool(m_sName, "ShowProgress",true);
	m_bSilent			= GetPrefixedRCBool(m_sName, "Silent",		false);
	//m_bUseSSL			= GetPrefixedRCBool(m_sName, "UseHTTPS",	false);

	return true;
}

////////////////////////////////////////////////////////////////////////////
//
// This function parses HTML file and generates output string
// 
bool CNews::ReadFromFile(const char *szFile, bool bFailIfNotExist)
{
	bool bFailed = false;
	m_bUpdated = false;

	HANDLE f = CreateFile(szFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (f == INVALID_HANDLE_VALUE)
	{
		if (bFailIfNotExist)
		{
			ErrorBox("Internal error: File successfully downloaded but cannot be open", m_bSilent);
		}
		return false;
	}
	DWORD file_length = GetFileSize(f, NULL);
	char *buf = new char[file_length+1];

	// let's start
	try
	{
		DWORD rd;
		if (!ReadFile(f, buf, file_length, &rd, NULL) || rd != file_length)
			throw "Internal error: File successfully opened by cannot be read";

		// first we need to load parameters from the HTML
		buf[file_length] = 0;

		m_params.clear();

		char errmsg[MAX_LINE_LENGTH];
		if (!ReadParameters(buf, file_length, &m_params, errmsg))
		{
			m_params.clear();
			throw errmsg;
		}

		m_bUpdated = false;
		if (m_sOutputFile != "")
		{
			string sOldTextForFile = m_sTextForFile;
			m_sTextForFile = "";

			StringVector params_backup = m_params;

			StringVector::iterator si = m_params.begin(), send = m_params.end();
			for( ; si != send; si++)
			{
				if (m_sNewFileLine != "\\n")
				{
					*si = str_replace(*si, "\r\n", m_sNewFileLine.c_str());
					*si = str_replace(*si, "\n", m_sNewFileLine.c_str());
					*si = str_replace(*si, "\r", m_sNewFileLine.c_str());
				}
			}
			if (!GenerateText(&m_params, m_sOutputString, &m_sTextForFile, false))
			{
				m_params.clear();
				throw m_sTextForFile.c_str();
			}
			m_bUpdated = (m_sTextForFile != sOldTextForFile);

			m_params = params_backup;
		}
		if (m_sOutputLabel != "")
		{
			string sOldText = m_sText;
			m_sText = "";
			StringVector::iterator si = m_params.begin(), send = m_params.end();
			for( ; si != send; si++)
			{
				if (m_sInputNewLineAs != "\\n")
				{
					*si = str_replace(*si, "\r\n",	m_sInputNewLineAs.c_str());
					*si = str_replace(*si, "\n",	m_sInputNewLineAs.c_str());
					*si = str_replace(*si, "\r",	m_sInputNewLineAs.c_str());
				}
				DeHTML(si, m_sNewLine);
				*si = str_replace(*si, "[", "[[");
				*si = str_replace(*si, "]", "]]");
				*si = str_replace(*si, ";", "[semicolon]");
			}

			if (!GenerateText(&m_params, m_sOutputString, &m_sText, true))
			{
				m_params.clear();
				throw m_sText.c_str();
			}
			m_bUpdated = m_bUpdated || (m_sText != sOldText);
		}
	} catch (const char *msg)
	{
		// one of our error messages
		char errmsg[MAX_LINE_LENGTH];
		StringCchPrintf(errmsg, MAX_LINE_LENGTH, "The following error occured during web-page parsing:\n%s", msg);
		ErrorBox(errmsg, m_bSilent);
		bFailed = true;
	}
	catch (...)
	{
		// unspecified exception
		ErrorBox("Unknown error occured during web-page parsing.", m_bSilent);
		bFailed = true;
	}
	// let's stop :)

	delete buf;

	// just close file handle, caller would erase file by itself
	CloseHandle(f);
	return !bFailed;
}


////////////////////////////////////////////////////////////////////////////
//
// Read paramteres from HTML using user-specified pattern
// 
bool CNews::ReadParameters(char *szString, int len, StringVector *v, char *errmsg)
{
	char *szBuf = new char[m_sInputString.length()+1];
	StringCchCopy(szBuf, MAX_LINE_LENGTH, m_sInputString.c_str());
	char *szPattern = szBuf;
	int i = 0;	// position of the cursor in the pattern
	int j = 0;	//					and in the HTML file

	while (*szPattern && *szString)
	{
		if (strnicmp(szPattern, "{%}",3) == 0)	// read parameter
		{
			szPattern += 3;
			char *f = FindFirstEscapeSequence(szPattern);
			if (!f)
			{
				f = szPattern + strlen(szPattern);
			} else if (f == szPattern)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "No escape-sequences can follow {%} tag");
				return false;
			}
			bool bFound = false;
			for (char *p = szString; *p; p++)
			{
				if (strnicmp(p, szPattern, f-szPattern) == 0)
				{
					bFound = true;
					break;
				}
			}
			if (!bFound)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "User-specified substring not found");
				return false;
			}
			char *szTmp = new char[p-szString+1];
			strncpy(szTmp, szString, p-szString);
			szTmp[p-szString] = 0;
			v->push_back((string)szTmp);

			delete szTmp;

			szString = p + (f - szPattern);
			szPattern = f;
		} else
		if (strnicmp(szPattern, "{%,",3) == 0)	// read parameter
		{
			szPattern+=3;
			char *err;
			int k = (int)strtod(szPattern, &err);
			if (k<1 || *err != '}')
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "{%,N} tag found but N is incorrect");
				return false;
			}
			if (strlen(szString) < (size_t)k)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "Parameter must be longer than existing remainder");
				return false;
			}
			char *szTmp = new char[k+1];
			strncpy(szTmp, szString, k);
			szTmp[k] = 0;
			v->push_back((string)szTmp);

			delete szTmp;

			szPattern = strchr(szPattern, '}') + 1;
			szString+=k;
		} else		// skip some symbols
		{
			bool bWithTag = false;
			if (strnicmp(szPattern, "{*}",3) == 0)
			{
				szPattern += 3;
				bWithTag = true;
			}
			char *f = FindFirstEscapeSequence(szPattern);
			if (!f)
			{
				delete szBuf;
				if (bWithTag)
					StringCchCopy(errmsg, MAX_LINE_LENGTH, "{*} is last escape-sequence in the string, so it cannot be applied");
				else
					StringCchCopy(errmsg, MAX_LINE_LENGTH, "Pusycat!");
				return false;
			}
			if (bWithTag && f == szPattern)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "No escape-sequences can follow {*} tag");
				return false;
			}
			bool bFound = false;
			for (char *p = szString; *p; p++)
			{
				if (strnicmp(p, szPattern, f-szPattern) == 0)
				{
					bFound = true;
					break;
				}
			}
			if (!bFound)
			{
				delete szBuf;
				StringCchCopy(errmsg, MAX_LINE_LENGTH, "User-specified substring not found");
				return false;
			}
			szString = p + (f - szPattern);
			szPattern = f;
		}
	}
	delete szBuf;
	return true;
}


////////////////////////////////////////////////////////////////////////////
//
// Enable autoupate
// 
void CNews::Enable()
{
	if (m_bActive) return;
	m_bActive = true;
	LastEventIDValue++;
	m_nIDEvent = LastEventIDValue;
	m_uTimerID = SetTimer(messageHandler, m_nIDEvent, 1000*m_UpdateInterval, NULL);
}

////////////////////////////////////////////////////////////////////////////
//
// Disable autoupdate
// 
void CNews::Disable()
{
	if (!m_bActive) return;
	m_bActive = false;
	KillTimer(messageHandler, m_uTimerID);
}

////////////////////////////////////////////////////////////////////////////
//
// Toggle autoupdate
// 
void CNews::Toggle()
{
	m_bActive ? Disable() : Enable();
}

////////////////////////////////////////////////////////////////////////////
//
// Change input pattern
// 
void CNews::SetInputString(const char *str)
{
	m_sInputString = str_replace((string)str, "{%quote}", "\"");
	m_sInputString = str_replace(m_sInputString, "{%dollar}", "$");
}

////////////////////////////////////////////////////////////////////////////
//
// Change output pattern
// 
void CNews::SetOutputString(const char *str)
{
	m_sOutputString = str_replace((string)str, "{%quote}", "\"");
	m_sOutputString = str_replace(m_sOutputString, "{%dollar}", "$");
}

////////////////////////////////////////////////////////////////////////////
//
// Check the news
// 
void CNews::Update()
{
	// if we are not on updating
	if (m_UpdateThread == 0)
	{
		m_bAbort = false;
		_beginthreadex(NULL, 0, UpdateNews, this, 0, &m_UpdateThread);
	}
}

////////////////////////////////////////////////////////////////////////////
//
// Set label text using the pattern specified by user
// 
void CNews::SetLabelText(const char *szLabelName, const char *szPattern)
{
	string sText;
	GenerateText(&m_params, (string)szPattern, &sText, true);
	::SetLabelText((string)szLabelName, sText);
}

////////////////////////////////////////////////////////////////////////////
//
// Replace {%N} with string values
// 
void CNews::ParseCommand(const char *szBang)
{
	string sText;
	string sBang = str_replace((string)szBang, "{%quote}", "\"");
	sBang = str_replace(sBang, "{%dollar}", "$");
	GenerateText(&m_params, sBang, &sText, true);
	for (size_t i = 0; i < sText.length(); i++)
	{
		if (sText[i] == '\"') sText[i] = '\'';
	}
	LSExecute(NULL, sText.c_str(), SW_SHOWDEFAULT);
}


////////////////////////////////////////////////////////////////////////////
//
// Show downloading progress
// 
void CNews::ShowProgress(DWORD nDataAvailable, DWORD nDownloaded)
{
	char szDownloaded[50], szFileLength[50], szPercentage[50];
	string s;
	StringCchPrintf(szDownloaded, 50, "%d", nDownloaded);
	if (nDataAvailable > 0)
	{
		StringCchPrintf(szFileLength, 50, "%d", nDataAvailable);
		StringCchPrintf(szPercentage, 50, "%d", 100*nDownloaded/nDataAvailable);
	}
	else
	{
		StringCchPrintf(szFileLength, 50, "<unknown>");
		StringCchPrintf(szPercentage, 50, "<unknown>");
	}
	s = str_replace(m_sProgress, "{downloaded}", szDownloaded);
	s = str_replace(s, "{size}", szFileLength);
	s = str_replace(s, "{%}", szPercentage);
	::SetLabelText(m_sOutputLabel, s);
}

////////////////////////////////////////////////////////////////////////////
//
// Download web page specified by URL to the local file
// 
bool CNews::URLDownloadToFile(const char *szFile)
{
	DWORD_PTR dwContext = NULL;
	char buf[1024];
	DWORD rd = 1024;
	DWORD nDownloaded = 0, nDataAvailable;
	bool bSuccess = false;

	HINTERNET hConnect = NULL, hRequest = NULL;

	HINTERNET hInternet = InternetOpen(NULL, PRE_CONFIG_INTERNET_ACCESS, NULL, NULL, 0);
	if (!m_bAbort && hInternet)
	{
		/*
		if (m_bUseSSL)
		{
			char szServer
			URL_COMPONENTS url_components;
			ZeroMemory(&url_components, sizeof(url_components));
			url_components.dwStructSize = sizeof(URL_COMPONENTS);
			InternetCrackUrl(m_sURL.c_str(), m_sURL.length(), ICU_DECODE, &url_components);
			hConnect = InternetConnect(
				hInternet,
				szServer,
				INTERNET_DEFAULT_HTTPS_PORT,
				m_sUserName.c_str(),
				m_sPassword.c_str(),
				INTERNET_SERVICE_HTTP,
				0, 0);
			HINTERNET hRequest = HttpOpenRequest (
				hConnect,                   // InternetConnect handle
				"GET",                      // Method
				"",                         // Object name
				HTTP_VERSION,               // Version
				"",                         // Referrer
				NULL,                       // Extra headers
				INTERNET_FLAG_RELOAD | INTERNET_FLAG_SECURE | INTERNET_FLAG_IGNORE_CERT_CN_INVALID | INTERNET_FLAG_IGNORE_CERT_DATE_INVALID,
				0);
		}
		else
		{
		*/
			HINTERNET hFile = InternetOpenUrl(
				hInternet,
				m_sURL.c_str(),
				NULL,
				0,
				INTERNET_FLAG_TRANSFER_BINARY | INTERNET_FLAG_RELOAD,
				dwContext);
		//}
		if (!m_bAbort && hFile)
		{
			HANDLE f = CreateFile(
				szFile,
				GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);
			if (!m_bAbort && f != INVALID_HANDLE_VALUE)
			{
				nDataAvailable = InternetSetFilePointer(hFile, 0, NULL, FILE_END, 0);
				InternetSetFilePointer(hFile, 0, NULL, FILE_BEGIN, 0);
				bSuccess = true;
				if (m_bShowProgress) ShowProgress(nDataAvailable, nDownloaded);
				while(InternetReadFile(hFile, buf, 1024, &rd) && rd)
				{
					DWORD wr;
					WriteFile(f, buf, rd, &wr, NULL);
					if (m_bAbort || wr != rd)
					{
						bSuccess = false;
						break;
					}
					nDownloaded += rd;
					if (m_bShowProgress) ShowProgress(nDataAvailable, nDownloaded);
				}
				CloseHandle(f);
			}
			InternetCloseHandle(hFile);
		}
		InternetCloseHandle(hInternet);
	}
	if (bSuccess && m_sLocalFile != "")
	{
		bSuccess = (CopyFile(szFile, m_sLocalFile.c_str(), FALSE) != FALSE);
	}
	return bSuccess;
}

////////////////////////////////////////////////////////////////////////////
//
// This thread checks news.
// We cannot download file from the main thread cause it can freeze LS
//
unsigned __stdcall UpdateNews(void* lpParameter)
{
	CNews *news = (CNews*)lpParameter;
	char szTempFolder[MAX_LINE_LENGTH] = "", szTempFileName[MAX_LINE_LENGTH] = "";
	try
	{
		GetTempPath(MAX_LINE_LENGTH, szTempFolder);
		GetTempFileName(szTempFolder, "lsw", 0, szTempFileName);
		if (!news->URLDownloadToFile(szTempFileName))
		{
			throw "There is no file with specified URL";
		}
		bool bParsed = news->ReadFromFile(szTempFileName, true);
		DeleteFile(szTempFileName);
		if (!bParsed)
		{
			throw "News file parsing failed";
		}
		if (news->m_sOnChecked != "" &&
			stricmp(news->m_sOnChecked.c_str(), "!none") != 0)
		{
			LSExecute(NULL, news->m_sOnChecked.c_str(), SW_SHOWNORMAL);
			if (news->m_bUpdated &&
				stricmp(news->m_sOnUpdated.c_str(), "!none") != 0 &&
				news->m_sOnUpdated != "" &&
				news->m_sLocalFile != "")
			{
				news->m_bUpdated = false;
				LSExecute(NULL, news->m_sOnUpdated.c_str(), SW_SHOWNORMAL);
			}
		}
		if (news->m_sOutputLabel != "")
		{
			SetLabelText(news->m_sOutputLabel, news->m_sText);
		}
		if (news->m_sOutputFile != "")
		{
			HANDLE f = CreateFile(
				news->m_sOutputFile.c_str(),
				GENERIC_WRITE,
				FILE_SHARE_READ,
				NULL,
				CREATE_ALWAYS,
				FILE_ATTRIBUTE_NORMAL,
				NULL);
			if (f != INVALID_HANDLE_VALUE)
			{
				DWORD wr;
				WriteFile(f, news->m_sTextForFile.c_str(), news->m_sTextForFile.length(), &wr, NULL);
				CloseHandle(f);
			}
		}
	}
	catch(const char *)
	{
		if (!news->m_bAbort && stricmp(news->m_sOnFailure.c_str(), "!none") != 0)
		{
			LSExecute(NULL, news->m_sOnFailure.c_str(), SW_SHOWNORMAL);
		}
	}
	catch(...)
	{
		ErrorBox("Internal error: Unknown exception raised", news->m_bSilent);
	}
	news->m_UpdateThread = 0;
	return 1;
}

